// Copyright 2003 Regents of the University of California

// SETI_BOINC is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2, or (at your option) any later
// version.

// SETI_BOINC is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
// more details.

// You should have received a copy of the GNU General Public License along
// with SETI_BOINC; see the file COPYING.  If not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

// analyzeFuncs.C
// $Id: analyzeFuncs.cpp,v 1.31 2004/12/23 19:07:08 korpela Exp $
//

#define DO_SMOOTH

#include "config.h"

//
// Modified to utilize IPP library, 5.0beta
//
//  see http://intel.com/software/products/ipp/index.htm
//
//  by Tetsuji Maverick Rai


#if defined(T7)
#pragma message("Optimized for P4 w/ SSE3")
#include "ipp_t7.h"    // P4 w SSE3
#elif defined(W7) || defined(M)
#pragma message("Optimized for P4 w/o SSE3")
#include "ipp_w7.h"    // P4 w/o SSE3 or M
#elif defined(A6)
#pragma message("Optimized for P3")
#include "ipp_a6.h"    // P3
#else
#pragma message("Others")
#include "ipp_px.h"    // others
#endif

#include "ipp.h"

// end of ipp headers


#include <stdio.h>
#include <iostream>
#include <stdlib.h>
#include <math.h>
#include <memory.h>

#include "util.h"
#include "s_util.h"

//#include "fft8g.h"  // we don't need this any longer
#include "seti.h"
#include "analyze.h"
#include "analyzeReport.h"
#include "gaussfit.h"
#include "spike.h"
#include "malloc_a.h"
#include "analyzeFuncs.h"
#include "analyzePoT.h"
#include "chirpfft.h"
#include "worker.h"
#include "filesys.h"
#ifdef BOINC_APP_GRAPHICS
#include "gdata.h"
#include "reduce.h"
#include "graphics_data.h"
#include "sah_gfx.h"
#endif

#include "boinc_api.h"

#ifndef M_PI
#define M_PI 3.14159265358979323846
#endif

ChirpFftPair_t* ChirpFftPairs = NULL;

double ProgressUnitSize;
double progress;

// Workspace for ippsFFT funcs

static int MaxBufSize = 0;
static Ipp8u* IppBuf = NULL;  // This will never be freed

// Workspace in v_BaseLineSmooth    Tetsuji Maverick Rai
#define MAXNRD 19     // length <= 2^(MAXNRD-1)
static IppsFFTSpec_C_32fc* BLSFftSpec[MAXNRD-2]; // FftSpec for BaseLineSmooth
static float* BLSPowerSpectrum;

// Initialization for v_BaseLineSmooth()

static void v_InitBaseLineSmooth(){
  int i;
  int BufSize = 0;

  for (i = 0; i< MAXNRD-3; i++){
    // here, faster the better
    ippsFFTInitAlloc_C_32fc(&BLSFftSpec[i], i+3, IPP_FFT_NODIV_BY_ANY,
			    ippAlgHintFast);
    ippsFFTGetBufSize_C_32fc(BLSFftSpec[i], &BufSize);
    if (MaxBufSize < BufSize){
      MaxBufSize = BufSize;
    }
  }
  if (IppBuf) ippsFree(IppBuf);
  IppBuf = ippsMalloc_8u(MaxBufSize);
  if(IppBuf == NULL) SETIERROR(MALLOC_FAILED, "IppBuf == NULL");

  BLSPowerSpectrum =
    (float*) malloc_a((1L << (MAXNRD-1))* sizeof(float),MEM_ALIGN);
  if (BLSPowerSpectrum == NULL) {
    printf("Could not allocate Power Spectrum array in \
v_InitBaseLineSmooth()\n");
    exit(1);
  }
}

// Initialization for v_BaseLineSmooth()

static void v_FreeBaseLineSmooth(){
  int i;
  int BufSize = 0;

  for (i = 0; i< MAXNRD-3; i++){
    // here, faster the better
    ippsFFTFree_C_32fc(BLSFftSpec[i]);
  }
  free_a(BLSPowerSpectrum); BLSPowerSpectrum = NULL;
}


// The main analysis function.  Args:
// state pointer to data, # of points, starting chirp/fftlen
//  Must be called with unchirped data;
//  this function modifies (chirps) the data in place
// swi  parsed WU header

int seti_analyze (ANALYSIS_STATE& state) {
  float* DataIn = state.data;
  int NumDataPoints = state.npoints;
  float* ChirpedData = NULL;
  float* WorkData = NULL;
  float* PowerSpectrum = NULL;
  int num_cfft                  = 0;
  int cfft_left;
  double MinChirpStep            = 0.0;
  float last_chirp = 0, chirp, chirprate;
  double progress_diff, progress_in_cfft;
  int retval=0;
  int icfft;
  int diffChirpCount = 0;
  int NumFfts, ifft, fftlen;
  int CurrentSub;
  int FftNum;
  unsigned long bitfield=swi.analysis_cfg.analysis_fft_lengths;
  unsigned long FftLen;
  unsigned long MaxFftLen = 0;

  //
  // spell for IPP
  // 
  //   Tetsuji Maverick Rai
  ippStaticInit();

  // writeing remarks
#ifdef PGO
  fprintf(stderr, "\
(Maniacally optimized with Intel C++ compiler 8.1/IPP library 4.1 \n\
by Tetsuji Maverick Rai, rev.08 PGO, " __TIMESTAMP__ ")\n");
#else
  fprintf(stderr, "\
(Maniacally optimized with Intel C++ compiler 8.1/IPP library 4.1 \n\
by Tetsuji Maverick Rai, rev.08, " __TIMESTAMP__ ")\n");
#endif
  //    for IPP libray we need this
  IppsFFTSpec_C_32fc* FftSpec[MAX_NUM_FFTS];
  int BufSize;

  // Allocate data array and work area arrays.

  ChirpedData = state.data;
  PowerSpectrum = (float*) calloc_a(NumDataPoints, sizeof(float), MEM_ALIGN);
  if(PowerSpectrum == NULL) SETIERROR(MALLOC_FAILED, "PowerSpectrum == NULL");

  v_InitBaseLineSmooth();  // Tetsuji Maverick Rai

  FftNum=0;
  FftLen=1;
  while (bitfield != 0) {
    if (bitfield & 1) {
      swi.analysis_fft_lengths[FftNum]=FftLen;
      FftNum++;
    }
    FftLen*=2;
    bitfield>>=1;
  }
  swi.num_fft_lengths=FftNum;

  for(FftNum = 0; FftNum < swi.num_fft_lengths; FftNum++) {
      // This is an important preparation for ipp fft function.
      // Tetsuji Maverick Rai
      int order = -1;
      if (MaxFftLen < swi.analysis_fft_lengths[FftNum])
	MaxFftLen = swi.analysis_fft_lengths[FftNum];
      for (int tmp = swi.analysis_fft_lengths[FftNum]; tmp; order++)
	tmp >>= 1;
      ippsFFTInitAlloc_C_32fc(&FftSpec[FftNum], order, IPP_FFT_NODIV_BY_ANY,
			      ippAlgHintFast);   // the faster the better 
      ippsFFTGetBufSize_C_32fc(FftSpec[FftNum], &BufSize);
      if (BufSize > MaxBufSize)  // sometimes BufSize == 0
	MaxBufSize = BufSize;
  }
  if (MaxBufSize){
    if (IppBuf)ippsFree(IppBuf);
    IppBuf = ippsMalloc_8u(MaxBufSize);
    if(IppBuf == NULL) SETIERROR(MALLOC_FAILED, "IppBuf == NULL");
  }

  // Smooth Baseline

#ifdef DO_SMOOTH
#ifdef BOINC_APP_GRAPHICS
  if (!nographics) strcpy(sah_graphics.status, "Doing Baseline Smoothing");
#endif
  retval = v_BaseLineSmooth(
             DataIn, NumDataPoints, swi.analysis_cfg.bsmooth_boxcar_length, 
             swi.analysis_cfg.bsmooth_chunk_size
           );
  if (retval) SETIERROR(retval,"from v_BaseLineSmooth");
#endif

  // gernerate table of chirp/fft pairs (we may read table from file if testing)
  if(cfft_file != NULL)
    num_cfft = ReadCFftFile(&ChirpFftPairs, &MinChirpStep);
  else
    num_cfft = GenChirpFftPairs(&ChirpFftPairs, &MinChirpStep);
  if(num_cfft == MALLOC_FAILED) {
    SETIERROR(MALLOC_FAILED, "num_cfft == MALLOC_FAILED");
  }

  // Get together various values that we'll need to analyse power over time
  ComputePoTInfo(num_cfft, NumDataPoints);

  // used to calculate percent done
  ProgressUnitSize = GetProgressUnitSize(NumDataPoints, num_cfft);

#ifdef DUMP_CHIRP
  fprintf(stderr, "size  = %d MinChirpStep = %f\n", num_cfft, MinChirpStep);
  for(icfft = 0; icfft < num_cfft; icfft++) {
    fprintf(stderr,"%f  %d\n", ChirpFftPairs[icfft].ChirpRate,
            ChirpFftPairs[icfft].FftLen);
  }
#endif

  // WorkData must be allocated outside loop (the largest one)
  WorkData = (float*)calloc_a(MaxFftLen*2, sizeof(float), MEM_ALIGN);
  if (WorkData == NULL) {
    SETIERROR(MALLOC_FAILED, "WorkData == NULL");
  }

  fflush (stderr);    // flush warnings which happened until now.

  // Loop through chirp/fft pairs - this is the top level analysis loop.
  for (icfft = state.icfft; icfft < num_cfft; icfft++) {
    fftlen    = ChirpFftPairs[icfft].FftLen;
    chirprate = ChirpFftPairs[icfft].ChirpRate;
    chirp     = ChirpFftPairs[icfft].ChirpRate - last_chirp;  // default = chirp _step_

    // avoid redundant chirping
    if(!VeryClose(chirp, 0.0, MinChirpStep/100.0)) {
      if( diffChirpCount % 1000 == 0 && diffChirpCount != 0 ) { // reload WU every 1000 chirps, but not the first time through
        if( swi.data_type == DATA_ENCODED ) {
          bits_to_floats(state.savedWUData, DataIn, state.npoints);
#ifdef DO_SMOOTH
#ifdef BOINC_APP_GRAPHICS
          if (!nographics) strcpy(sah_graphics.status, "Doing Baseline Smoothing");
#endif
          retval = v_BaseLineSmooth(
                   DataIn, NumDataPoints, swi.analysis_cfg.bsmooth_boxcar_length, 
                   swi.analysis_cfg.bsmooth_chunk_size
                   );
#endif
          chirp = chirprate;  // change to absolute chirp _rate_
        }
      }

#ifdef BOINC_APP_GRAPHICS
      if (!nographics) strcpy(sah_graphics.status, "Chirping data");
#endif

      retval = v_ChirpData(
                 DataIn,
                 ChirpedData,
                 chirp,
                 NumDataPoints,
                 swi.subband_sample_rate
               );

      if (retval) SETIERROR(retval, "from v_ChirpData()");

      progress += (float)(ProgressUnitSize * CHIRP_PROGRESS_FACTOR);
      diffChirpCount++;
    }

    last_chirp = chirprate;

    // Process this FFT length.
    // For a given FFT length (at a given chirp), we construct
    // PowerSpectrum[] which is a "waterfall" array of power spectra,
    // each fftlen long on the frequency axis and sample time long
    // on the time axis.
    // As we go along, we check each spectra for spikes.

    state.icfft = icfft;     // update analysis state

    // Find index into FFT length table for the current
    // FFT length.  This will be the same index needed
    // for ooura's coeffecient and bit reverse tables.
    for(FftNum = 0; FftNum < swi.num_fft_lengths; FftNum++) {
      if(swi.analysis_fft_lengths[FftNum] == fftlen) {
        break;
      }
    }

#ifdef BOINC_APP_GRAPHICS
      if (!nographics) {
        sah_graphics.fft_info.chirp_rate = chirprate;
        sah_graphics.fft_info.fft_len = fftlen;
        strcpy(sah_graphics.status, "Computing Fast Fourier Transform");
      }
#endif

    // If PoT freq bin is non-negative, we are into PoT analysis
    // for this cfft pair and should not re-output an "ogh" line.
    if(state.PoT_freq_bin == -1) {
      retval = result_group_start();
      if (retval) SETIERROR(retval,"from result_group_start");
    }

    // Number of FFTs for this length
    NumFfts   = NumDataPoints / fftlen;

    // WorkData was allocated here in the original code, but now it's allocated
    // elsewhere

#ifdef BOINC_APP_GRAPHICS
    GRAPH_BUFFER* gbp;
    if (!nographics) {
      gbp = sah_graphics.get_generate_buffer(true);
      if (gbp) {
        gbp->rarray.init_data(fftlen, NumFfts);
      }
    }
#endif
    for (ifft = 0; ifft < NumFfts; ifft++) {

      CurrentSub = fftlen * 2 * ifft;

      //
      // Now it's high time to use ipp fft
      //
      //  Tetsuji Maverick Rai
      ippsFFTInv_CToC_32fc((Ipp32fc*)&ChirpedData[CurrentSub],
			   (Ipp32fc*)WorkData,
			   FftSpec[FftNum], IppBuf);


      // replace freq with power
      v_GetPowerSpectrum( WorkData,
                          &PowerSpectrum[CurrentSub / 2],
                          fftlen
                        );

      // any ETIs ?!
      // If PoT freq bin is non-negative, we are into PoT analysis
      // for this cfft pair and need not redo spike finding.
      if(state.PoT_freq_bin == -1) {
        retval = FindSpikes(
                   &PowerSpectrum[CurrentSub / 2],
                   fftlen,
                   ifft,
                   swi
                 );
        if (retval) SETIERROR(retval,"from FindSpikes");
      }

      //progress = ((float)icfft)/num_cfft + ((float)ifft)/(NumFfts*num_cfft);
      progress += (float)((ProgressUnitSize*SPIKE_PROGRESS_FACTOR) / NumFfts);
      if (progress > 1.0f) progress = 1.0f; // prevent display of > 100%
#ifdef BOINC_APP_GRAPHICS
      if (!nographics) { 
        if (gbp) {
          gbp->rarray.add_source_row(PowerSpectrum+fftlen*ifft);
        }
        sah_graphics.local_progress = (((float)ifft+1)/NumFfts);
      }
#endif
      boinc_fraction_done(progress);
// jeffc
//fprintf(stderr, "S fft len %d  progress = %12.10f\n", fftlen, progress);
    } // loop through chirped data array

#ifdef BOINC_APP_GRAPHICS
    if (!nographics) {
      if (gbp) {
        sah_graphics.generate_done(gbp);
      }
    }
#endif
    boinc_fraction_done(progress);
// jeffc
//fprintf(stderr, "Sdone fft len %d  progress = %12.10f\n", fftlen, progress);


    // Analyze Power over Time.  May return quickly if this FFT
    // length and/or this WUs slew rate places the data block
    // outside PoT analysis limits.
    retval = analyze_pot(PowerSpectrum, NumDataPoints, fftlen);
    if (retval) SETIERROR(retval,"from analyze_pot");

#ifdef BOINC_APP_GRAPHICS
    // switch the display back to "best of" signals
    //
    if (!nographics) {
      sah_graphics.gi.copy(best_gauss, true);
      sah_graphics.pi.copy(best_pulse, true);
      sah_graphics.ti.copy(best_triplet, true);
    }
#endif
    // Force progress to 100% before calling result_group_end() to store
    //  100% in state file so it will survive exit & relaunch
    if (icfft == (num_cfft-1)) {
      progress = 1;
      boinc_fraction_done(progress);
// jeffc
//fprintf(stderr, "DONE fft len %d  progress = %12.10f\n", fftlen, progress);

    }
    retval = checkpoint();
    if (retval) SETIERROR(retval,"from checkpoint() in seti_analyse()");

    // Progress adjustment - start checking late in the game (90%).
    // If progress_in_cfft is more than 1 cfft pair ahead of
    // progress, add the diff divided by the number of cfft's we have left.
    if((progress_in_cfft = (float)icfft/(float)num_cfft) > 0.9) {
      progress_diff = progress_in_cfft - progress;
      if(progress_diff > 1.0f/(float)num_cfft) {
        cfft_left = num_cfft-icfft > 0 ? num_cfft-icfft : 1;
        progress += progress_diff/cfft_left;
      }
    }

  } // loop over chirp/fftlen pairs

  // Return the "best of" signals.  This may include duplicates of
  // already reported interesting signals.
  if(best_spike->score) {
    	retval = outfile.printf("%s", best_spike->s.print_xml(0,0,1,"best_spike").c_str());
	if (retval < 0) {
  		SETIERROR(WRITE_FAILED,"from outfile.printf (best spike) in seti_analyze()");
	}

  }
  if(best_gauss->score) {
    	retval = outfile.printf("%s", best_gauss->g.print_xml(0,0,1,"best_gaussian").c_str());
	if (retval < 0) {
  		SETIERROR(WRITE_FAILED,"from outfile.printf (best gaussian) in seti_analyze()");
	}
  }
  if(best_pulse->score) {
    	retval = outfile.printf("%s", best_pulse->p.print_xml(0,0,1,"best_pulse").c_str());
	if (retval < 0) {
  		SETIERROR(WRITE_FAILED,"from outfile.printf (best pulse) in seti_analyze()");
	}
  }
  if(best_triplet->score) {
    	retval = outfile.printf("%s", best_triplet->t.print_xml(0,0,1,"best_triplet").c_str());
	if (retval < 0) {
  		SETIERROR(WRITE_FAILED,"from outfile.printf (best triplet) in seti_analyze()");
	}
  }

#ifdef BOINC_APP_GRAPHICS
  if (!nographics) strcpy(sah_graphics.status, "Work unit done");
#endif

#if defined(_WIN32) && defined(_DEBUG)

  if (PowerSpectrum) _aligned_free(PowerSpectrum);
  for(FftNum = 0; FftNum < swi.num_fft_lengths; FftNum++) {
    if (BitRevTab[FftNum]) _aligned_free(BitRevTab[FftNum]);
  }
  for(FftNum = 0; FftNum < swi.num_fft_lengths; FftNum++) {
    if (CoeffTab[FftNum]) _aligned_free(CoeffTab[FftNum]);
  }
  if (WorkData) _aligned_free(WorkData);

#else

  if (PowerSpectrum) free_a(PowerSpectrum);
  // modified by Tetsuji Maverick Rai for ipp library
  for(FftNum = 0; FftNum < swi.num_fft_lengths; FftNum++) {
    if (FftSpec[FftNum]) ippsFFTFree_C_32fc(FftSpec[FftNum]);
  }
  // WorkData should be freed here, outside loop
  // Tetsuji Maverick Rai
  if (WorkData) free_a(WorkData);
  WorkData = NULL;

  v_FreeBaseLineSmooth();    // Free BLSFftSpec   Tetsuji Maverick Rai
#endif

  if (ChirpFftPairs) free(ChirpFftPairs);

  // jeffc
  //retval = outfile.flush();
  retval = checkpoint();		// force a final checkpoint
  xml_indent(-2);
  outfile.printf("</result>");
  outfile.close();
  //if (retval) SETIERROR(WRITE_FAILED,"from outfile.fflush in seti_analyze()");

  return retval;
}


int v_BaseLineSmooth(
  float* DataIn,
  int NumDataPoints,
  int BoxCarLength,
  int NumPointsInChunk
) {

  // We use a sliding boxcar method for baseline smoothing.  Input data
  // is the time domain.  It is transformed (using a separate array)
  // into the frequency domain.  After baseline smoothing is done
  // in the frequency domain, it is transformed back into the time domain.

  int h, i, j, k, i_minus_1;
  float* DataInChunk;
//  float* PowerSpectrum;
  float Total, LocalMean, ScaleFactor;
  int NumTimeChunks, TimeChunk, Endpoint;

  NumTimeChunks = NumDataPoints / NumPointsInChunk;

//  allocation of BLSPowerSpectrum is done in v_InitBaseLineSmooth()
//   Tetsuji Maverick Rai

  int order = -1;
  for (int tmp = NumPointsInChunk; tmp; order++)
    tmp >>= 1;

  for (TimeChunk = 0; TimeChunk < NumTimeChunks; TimeChunk++) {
#ifdef BOINC_APP_GRAPHICS
    if (!nographics) sah_graphics.local_progress = (((float)TimeChunk)/NumTimeChunks);
#endif

    DataInChunk = &DataIn[TimeChunk*NumPointsInChunk*2];

    // transform to freq
    // Here goes IPP transform (Inv)
    ippsFFTInv_CToC_32fc((Ipp32fc*)DataInChunk, (Ipp32fc*)DataInChunk,
			 BLSFftSpec[order-3], IppBuf);

    v_GetPowerSpectrum(
      DataInChunk, BLSPowerSpectrum, NumPointsInChunk
    );

    // Begin: normalize in freq. domain via sliding boxcar

    Endpoint = NumPointsInChunk / 2;

    i = Endpoint;         // start i at lowest negative freq;
    // this is low point in first boxcar
    j = i + BoxCarLength / 2;        // start j midpoint in first boxcar
    k = i + BoxCarLength;            // start k at end of first boxcar;
    // thus a boxcar is i----j----k
    Total = 0;
    for (h = i; h < k; h++) {            // Get mean for first boxcar
      Total += BLSPowerSpectrum[h];
    }
    LocalMean = Total / BoxCarLength;
    ScaleFactor = (float)(sqrt(LocalMean / 2));

    // normalize 1st half of 1st boxcar
    for (h = i; h < j; h++) {
      DataInChunk[h*2]   /= ScaleFactor;
      DataInChunk[h*2+1] /= ScaleFactor;
    }

    for (; k != Endpoint; i++, j++, k++) {  // sliding boxcar
      if (k == NumPointsInChunk) {    // take care of wrapping
        k = 0;
      }
      if (j == NumPointsInChunk) {
        j = 0;
      }
      if (i == NumPointsInChunk) {
        i = 0;
        i_minus_1 = NumPointsInChunk -1;
      } else {
        i_minus_1 = i - 1;
      }

      LocalMean = LocalMean
                  - BLSPowerSpectrum[i_minus_1] / BoxCarLength
                  + BLSPowerSpectrum[k] / BoxCarLength;
      ScaleFactor = (float)(sqrt(LocalMean / 2));

      DataInChunk[j*2]   /= ScaleFactor;
      DataInChunk[j*2+1] /= ScaleFactor;

    } // End sliding boxcar

    // normalize final half of final boxcar
    for (h = j; h < k; h++) {
      DataInChunk[h*2]     /= ScaleFactor;
      DataInChunk[h*2+1] /= ScaleFactor;
    }

    // End: normalize in freq. domain via sliding boxcar
    // transform back to time

    // Here we use IPP
    ippsFFTFwd_CToC_32fc((Ipp32fc*)DataInChunk, (Ipp32fc*)DataInChunk,
			 BLSFftSpec[order-3], IppBuf);

    // return powers to normal
    for (i = 0; i < NumPointsInChunk * 2; i++) {
      DataInChunk[i] /= NumPointsInChunk;
    }
  }

//  if (PowerSpectrum) free_a(PowerSpectrum);
  return 0;
}





void v_GetPowerSpectrum(
  float* FreqData,
  float* PowerSpectrum,
  int NumDataPoints
) {

#if !defined(_WIN32) &&(defined(T7) || defined(W7))   // use SSE/SSE2
#pragma message("Use hand-assembly code")
  //
  //  Hand-assembled by Tetsuji Maverick Rai
  //
  asm("\
        shr     $1,%2;  /* I'm sure it's an even number */ \
        test    %2,%2;  /* check NumDataPoints == 0 or not */ \
        jz      exit_asm;\
loop1:\
        movaps (%0),%%xmm0;       /* load 2 complexes at once */ \
        mulps  %%xmm0,%%xmm0;      \
        pshufd $0x31,%%xmm0,%%xmm1;  /* shift one fp to the right */ \
        addps  %%xmm1,%%xmm0;      /* abandon 2nd,4th fps */ \
        movss  %%xmm0,(%1);         /* store 1st result */ \
        pshufd $0x02,%%xmm0,%%xmm1;  /* shift two fps to the right */ \
        movss  %%xmm1,4(%1);        /* store 2nd result */\
        add    $16,%0; \
        add    $8,%1;  \
        dec    %2; \
        jnz    loop1; \
exit_asm:\
"
       : /* no outputs */
       : "r"(FreqData), "r"(PowerSpectrum), "r"(NumDataPoints)
       : "xmm0","xmm1"    // maybe not necessary
      );
#else

  int i, j;

#ifdef __ICC  // force vectorization here   Tetsuji Maverick Rai
  #pragma ivdep
  #pragma vector aligned
  __assume_aligned(FreqData,16);
#endif
  for (i = 0, j = 0; i < NumDataPoints * 2; i += 2, j++) {
    PowerSpectrum[j] = FreqData[i] * FreqData[i]
                       + FreqData[i+1] * FreqData[i+1];
  }
#endif

}

// chirp_rate is in Hz per second

#ifndef USE_INTEL_OPT_CODE
int v_ChirpData(
  float* fp_DataArray,
  float* fp_ChirpDataArray,
  float chirp_rate,
  int  ul_NumDataPoints,
  double sample_rate
) {
  int i, j;
  float c, d, real, imag;
  double time, ang;

  if (chirp_rate == 0.0) {
    memcpy(fp_ChirpDataArray,
           fp_DataArray,
           (int)ul_NumDataPoints * 2 * sizeof(float)
          );    // NOTE INT CAST
  } else {
    for (i = 0, j = 0; i < ul_NumDataPoints * 2; i += 2, j++) {
      time = j/sample_rate;
      ang = M_PI*2*chirp_rate*time*time;
      c = (float) cos(ang);
      d = (float) sin(ang);

      // Sometimes chirping is done in place.
      // We don't want to overwrite data prematurely.
      real = fp_DataArray[i] * c - fp_DataArray[i+1] * d;
      imag = fp_DataArray[i] * d + fp_DataArray[i+1] * c;
      fp_ChirpDataArray[i] = real;
      fp_ChirpDataArray[i+1] = imag;
    }
  }
  return 0;
}
#endif // USE_INTEL_OPT_CODE
